// Switch.cpp: implementation of the Switch class.
//
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "Switch.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
QueueHybrid::QueueHybrid()
{
  m_burstSize = 0.0;
  m_tokenRate = 0.0;
  m_id = 0;
  m_total = 0;
  m_outNum = 0;
  m_loss = 0;
  m_allTotal = 0;
  m_allOutNum = 0;
  m_allLoss = 0;
  m_threshold = 0;
  m_tick = 0;
  m_vClock = 0;
}

QueueHybrid::~QueueHybrid()
{
  ListPVCIt pVcIt;
  PVC pVc;

  for (pVcIt=m_listPVcPast.begin(); pVcIt!=m_listPVcPast.end(); pVcIt++)
  {
    pVc = *pVcIt;
    delete pVc;
  }
}

void QueueHybrid::Check(size_t &vcnum)
{
  DWORD loss, out, total;
  ListPVCIt pVcIt;
  PVC pVc;
  double rate, burst;

  loss=out=total=0;
  rate = 0.0;
  burst = 0;
  vcnum = 0;
  for (pVcIt=m_listPVc.begin(); pVcIt!=m_listPVc.end(); pVcIt++)
  {
    pVc = *pVcIt;
    loss += pVc->m_allLoss;
    out += pVc->m_allOutNum;
    total += pVc->m_allTotal;
    rate += 1.0/pVc->m_tokenInterval;
    burst += pVc->m_burstSize;
    if (pVc->m_flag == VC_STOP)
      assert(pVc->m_num>0);
    else
      vcnum++;
  }
  for (pVcIt=m_listPVcPast.begin(); pVcIt!=m_listPVcPast.end(); pVcIt++)
  {
    pVc = *pVcIt;
    loss += pVc->m_allLoss;
    out += pVc->m_allOutNum;
    total += pVc->m_allTotal;
  }
  verify(fabs(burst-m_burstSize)<1e-11);
  verify(fabs(rate-m_tokenRate)<1e-11);
  verify(total == m_allTotal);
  verify(total == loss+out+m_quePPacket.size());
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Switch::Switch(int switchType, double warmUp, double bufSize, double totalHeadroom, // MBytes
               DWORD outputInterval, int maxQueue, 
               double checkInterval, // seconds
               int method, char methodName[])
{
  /*
  // for debug
  for (int i=0; i<9; i++)
    m_maxLen[i]=0;
  m_max=0;
  */
  m_switchType = switchType;
  m_warmUp = INT64(warmUp * TimeScale);
  m_maxQueue = UINT_MAX; // for WFQ
  m_outputInterval = outputInterval;
  m_curTime = 0;
  m_prevTime = 0;
  m_checkTime = 0;
  m_interval = INT64(checkInterval * TimeScale);
  m_maxQueue = maxQueue;
  m_method = method;
  m_bufSize = DWORD(bufSize * 1e6 / CELL_SIZE);
  m_totalHeadroom = DWORD(totalHeadroom * 1e6 / CELL_SIZE);
  m_headroom = m_totalHeadroom;
  m_totalHoles = m_bufSize - m_totalHeadroom;
  m_holes = m_totalHoles;

  m_total = 0;
  m_outNum = 0;
  m_loss = 0;

  m_allTotal = 0;
  m_allOutNum = 0;
  m_allLoss = 0;

  /*
  m_pastTotal = 0;
  m_pastOutNum = 0;
  m_pastLoss = 0;

  m_pastAllTotal = 0;
  m_pastAllOutNum = 0;
  m_pastAllLoss = 0;
  */
  m_rand.srandom((m_bufSize+m_headroom)*(m_bufSize+m_headroom));
  m_curRRIndex = 0;
  switch (m_switchType)
  {
  case FIFO:
    {
      strcpy(m_name,"FIFO");
      break;
    }
  case FIFO_BUF:
    {
      strcpy(m_name,"FIFOBUF");
      break;
    }
  case WFQ:
    {
      strcpy(m_name,"WFQ");
      break;
    }
  case Hybrid:
    {
      size_t i;
      PQueueHybrid pQue;
      methodName[8]='\0';
      sprintf(m_name,"%-8s %-3lu", methodName, m_maxQueue);
      if (m_method==ADAPTIVE)
      {
        for (i=0; i<m_maxQueue; i++)
        {
          pQue = new QueueHybrid;
          pQue->m_id = i;
          m_vecPQueHybrid.push_back(pQue);
          pQue->m_vClock = UINT_MAX;
        }
      }
      break;
    }
  default:
    Error("Wrong switch type!");
  }
}

Switch::~Switch()
{
  VectorPQueueHybridIt it;
  PPacket pPacket;

  /*
  // for debug
  int i,j;
  cout << m_name << "\n";
  for (i=0; i<m_vecPQueHybrid.size(); i++)
  {
    cout << "MaxLen[" << i << "]=" << m_maxLen[i];
    for (j = 0; j<m_vecPQueHybrid.size(); j++)
      if (m_vecPQueHybrid[j]->m_id==i)
        break;
    assert(j!=m_vecPQueHybrid.size());
    cout << " Thresh[" << i << "]=" << m_vecPQueHybrid[j]->m_threshold << endl;
  }
  cout << endl << "Max=" << m_max << "  BufSize:" << m_bufSize << endl;
  */

  for (it=m_vecPQueHybrid.begin(); it != m_vecPQueHybrid.end(); it++)
  {
    while ((*it)->m_quePPacket.size()>0)
    {
      pPacket = (*it)->m_quePPacket.front();
      (*it)->m_quePPacket.pop_front();
      delete pPacket;
    }
    delete *it;
  }
}

// output a packet to the output link
void Switch::PacketOut(INT64 curTime)
{
  PQueueHybrid pQue;
  ListVCIt pVcIt;
  PVC pVc;

  assert(m_curTime<=curTime);
  m_curTime = curTime;
  if (m_vecPQueHybrid.size()==0)
    return;

  pop_heap(m_vecPQueHybrid.begin(), m_vecPQueHybrid.end(), LaterVirtualClock);
  pQue = m_vecPQueHybrid.back(); // get the queue with the smallest virtual clock
  
  if (pQue->m_quePPacket.empty()) // there should be no packets in switch
  {
    VectorPQueueHybridIt it;
    
    for (it=m_vecPQueHybrid.begin(); it != m_vecPQueHybrid.end(); it++)
    {
      assert((*it)->m_quePPacket.empty());
    }
    return;
  }

  pVc = OutputPacket(pQue);
  if (pVc!=NULL)
  {
    pVc->m_allOutNum++;
    pQue->m_allOutNum++;
    m_allOutNum++;

    if (m_curTime>m_warmUp)
    {
      pVc->m_outNum++;
      pQue->m_outNum++;
      m_outNum++;
      assert(m_outNum<0xFFFFFFFF);
    }
  }

  /*
  if (pQue->m_listPVc.size()==0) // erase this queue
  {
    // record information before delete the queue
    m_pastAllTotal += pQue->m_allTotal;
    m_pastAllOutNum += pQue->m_allOutNum;
    m_pastAllLoss += pQue->m_allLoss;
    m_pastTotal += pQue->m_total;
    m_pastOutNum += pQue->m_outNum;
    m_pastLoss += pQue->m_loss;

    delete pQue;
    m_vecPQueHybrid.pop_back();
  } 
  else 
  */
  if (!pQue->m_quePPacket.empty())
  { 
    if ((m_switchType == FIFO_BUF)||(m_switchType == FIFO))
      pQue->m_vClock = pQue->m_quePPacket.front()->m_time;
    else
      pQue->m_vClock += pQue->m_tick;
  }
  push_heap(m_vecPQueHybrid.begin(), m_vecPQueHybrid.end(), LaterVirtualClock);
}

// output a packet from pQue, return the VC it belongs to
PVC Switch::OutputPacket(PQueueHybrid pQue)
{
  ListVCIt pVcIt;
  ListPVCIt ppVcIt;
  PPacket pPacket;
  PVC pVc;

  pPacket = pQue->m_quePPacket.front(); // pPacket is outputed to the output link
  pQue->m_quePPacket.pop_front();

  // ofLog2 << *pPacket;

  pVcIt = m_vcHash.GetVcFromHash(pPacket->m_vcId); // pPacket comes from pVcIt
  pVc = &(*pVcIt);
  assert(int(pVc->m_id) == pPacket->m_vcId);
  pVc->m_num--;
  assert(pVc->m_num>=0);

  assert(m_prevTime<=pQue->m_vClock);
  assert(  !((m_switchType==FIFO)||(m_switchType==FIFO_BUF)) || 
            (pQue->m_vClock==pPacket->m_time));
  m_prevTime = pQue->m_vClock;

  //TRACE("Out(%d,%d)(T:%u,L:%u,O:%u):",pPacket->m_vcId,
  //          pPacket->m_seq,m_allTotal,m_allLoss,m_allOutNum);
  TraceAll();
  delete pPacket;

  if ((pVc->m_num==0) && (pVc->m_flag == VC_STOP)) // stop this vc now
  {
    ListPVCIt ppVcIt;

    for (ppVcIt=pQue->m_listPVc.begin(); ppVcIt!=pQue->m_listPVc.end(); ppVcIt++)
      if ((*ppVcIt)->m_id == pVc->m_id)
        break;
    assert(ppVcIt!=pQue->m_listPVc.end());
    pQue->m_listPVcPast.push_back(new VC(*pVc));
    pVc = pQue->m_listPVcPast.back();
    pQue->m_listPVc.erase(ppVcIt);
    if (pQue->m_listPVc.size()==0)
    {
      pQue->m_tokenRate = 0.0;
      pQue->m_burstSize = 0.0;
      pQue->m_tick = UINT_MAX;
    }
    else
    {
      pQue->m_burstSize -= pVc->m_burstSize;
      pQue->m_tokenRate -= 1.0/pVc->m_tokenInterval;
      pQue->m_tick = INT64(1/pQue->m_tokenRate);
    }
    m_vcHash.EraseVc(pVcIt);
    AdjustQueueThreshold();
  }

  m_headroom++;
  if (m_headroom > m_totalHeadroom)
  {
    m_holes++;
    m_headroom--;
    assert(m_headroom == m_totalHeadroom);
    assert(m_holes <= m_totalHoles);
  }
  return pVc;
}

// handle event, maybe a packet, virtual channel start or stop
// put pPacket to a queue
// allocate a new Packet from the parameter
void Switch::HandleEvent(PEvent pEvent)
{
  //ofLog << *(pEvent->pPacket);

  //CheckVClock(pEvent->pPacket->m_time);
  assert(m_curTime <= pEvent->pPacket->m_time);
  m_curTime = pEvent->pPacket->m_time;
  switch (pEvent->what)
  {
  case EVT_VC_START: // a new virtual channel starts
  {
    AcceptVc(pEvent);
    break;
  }
  case EVT_VC_STOP:   // a virtual channel stops
  {
    ListVCIt pVcIt;
    ListPVCIt ppVcIt;
    PQueueHybrid pQue;
    PVC pVc, pNewVc;

    pVcIt = m_vcHash.GetVcFromHash(pEvent->pPacket->m_vcId);
    pVcIt->m_flag = VC_STOP;

    if (pVcIt->m_num==0) // move this vc to m_listPVcPast
    {
      pQue = pVcIt->m_pQueHybrid;
      for (ppVcIt=pQue->m_listPVc.begin(); ppVcIt!=pQue->m_listPVc.end(); ppVcIt++)
      {
        pVc = *ppVcIt;
        if (pVc->m_id == pVcIt->m_id)
          break;
      }
    
      assert(ppVcIt!=pQue->m_listPVc.end());
      pNewVc = new VC(*pVc);
      pQue->m_listPVcPast.push_back(pNewVc);
      pQue->m_listPVc.erase(ppVcIt);
      if (pQue->m_listPVc.size()==0)
      {
        pQue->m_tokenRate = 0.0;
        pQue->m_burstSize = 0.0;
        pQue->m_tick = UINT_MAX;
      }
      else
      {
        pQue->m_burstSize -= pVc->m_burstSize;
        pQue->m_tokenRate -= 1.0/pVc->m_tokenInterval;
        pQue->m_tick = INT64(1/pQue->m_tokenRate);
      }
      m_vcHash.EraseVc(pVcIt);
      AdjustQueueThreshold();
    }
    break;
  }
  case EVT_NON_CONF: // this flow is non-conformant
  {
    break;
  }
  case EVT_PACKET: // a packet enters the switch
  {
    PPacket pPacket;
    PPacket pNewPacket;
    PQueueHybrid pQue;
    ListVCIt pVcIt;
    BOOL loss;
    int extra;

    assert(m_vecPQueHybrid.size()>0);
    pPacket = pEvent->pPacket;
    pVcIt = m_vcHash.GetVcFromHash(pPacket->m_vcId);
    pQue = pVcIt->m_pQueHybrid;
    assert((m_method!=PRE_SET) || (pVcIt->m_qid == pQue->m_id));

    if (m_allTotal==0) // only print at the first time
      PrintThreshold();
    pVcIt->m_allTotal++;
    pQue->m_allTotal++;
    m_allTotal++;

    if (m_curTime > m_warmUp)
    {
      m_total++;
      if (pQue->m_total==0)
        pQue->m_total = pQue->m_quePPacket.size();
      if (pVcIt->m_total==0) // the first time after warmUp
        pVcIt->m_total = pVcIt->m_num;
      pQue->m_total++;
      pVcIt->m_total++;
    }

    loss = FALSE;
    extra = pQue->m_quePPacket.size() - pQue->m_threshold;

    /*
    // for debug
    size_t i;
    for (i=0; i<m_vecPQueHybrid.size(); i++)
    {
      if (pQue->m_id == m_vecPQueHybrid[i]->m_id)
        break;
    }
    assert(i!=m_vecPQueHybrid.size());
    assert(pQue->m_quePPacket.size()==
           m_vecPQueHybrid[i]->m_quePPacket.size());

    if (m_maxLen[pQue->m_id]<pQue->m_quePPacket.size())
      m_maxLen[pQue->m_id] = pQue->m_quePPacket.size();

    if (m_total-m_loss-m_outNum>m_max)
      m_max = m_total-m_loss-m_outNum;
    */
    if (extra < 0)
    {
      if (m_holes>0)
        m_holes--;
      else if (m_headroom>0)
        m_headroom--;
      else
        loss = TRUE;
    }
    else
    {
      if ((m_holes>0) && 
          (extra+1 < int(m_holes-1)))
        m_holes--;
      else
        loss = TRUE;
    }

    if (loss) // drop this cell
    {
      /*
      if (pVcIt->m_conform)
      {
        ofLog << *(pEvent->pPacket);
        ofLog << "  QueSize:" << pQue->m_quePPacket.size()
              << " vc num:" << pVcIt->m_num << endl;
      }*/

      pVcIt->m_allLoss++;
      pQue->m_allLoss++;
      m_allLoss++;

      if (m_curTime > m_warmUp)
      {
        pVcIt->m_loss++;
        pQue->m_loss++;
        m_loss++;
      }
    }
    else // put the cell to the queue
    {
      pVcIt->m_num++;
      pNewPacket = new Packet(*pPacket);
      pQue->m_quePPacket.push_back(pNewPacket); 
      if (pQue->m_quePPacket.size()==1)
      {
        if ((m_switchType == FIFO_BUF)||(m_switchType == FIFO))
        {
          assert(pQue->m_vClock <= pQue->m_quePPacket.front()->m_time);
          pQue->m_vClock = pQue->m_quePPacket.front()->m_time;
        }
        else
        {
          pQue->m_vClock += pQue->m_tick;
          if (m_prevTime > pQue->m_vClock)
            m_prevTime = pQue->m_vClock; 
        }
        make_heap(m_vecPQueHybrid.begin(), m_vecPQueHybrid.end(), LaterVirtualClock);
        /*
        VectorPQueueHybridIt pVecQueIt;
        pVecQueIt = m_vecPQueHybrid.end()-1;
        while ((*pVecQueIt)->m_id != pQue->m_id)
          pVecQueIt--;
        push_heap(m_vecPQueHybrid.begin(), pVecQueIt+1, LaterVirtualClock);
        DebugPrint();
        */
      }
    }

    //TRACE("In(%d,%d)(T:%u,L:%u,O:%u):",pPacket->m_vcId,
    //      pPacket->m_seq,m_allTotal,m_allLoss,m_allOutNum);
    TraceAll();
    break;
  }
  default:
    assert(FALSE);
  }
}

void Switch::AcceptVc(PEvent pEvent)
{
  PVC pNewVc;
  PQueueHybrid pQue;

  if (m_switchType == FIFO)
  {
    if (m_vecPQueHybrid.size()<m_maxQueue)
    {
      assert(m_vecPQueHybrid.size()==0);
      pQue = new QueueHybrid;
      pQue->m_id = m_vecPQueHybrid.size();
      m_vecPQueHybrid.push_back(pQue);
      assert(m_vecPQueHybrid.size()!=0);
      pQue->m_vClock = pEvent->pPacket->m_time;
    }
    else
      pQue = m_vecPQueHybrid[0];
  }
  else if (m_method==NONE) // for FIFO_BUF and WFQ
  {
    assert(m_vecPQueHybrid.size()<m_maxQueue);
    pQue = new QueueHybrid;
    pQue->m_id = m_vecPQueHybrid.size();
    m_vecPQueHybrid.push_back(pQue);
    assert(m_vecPQueHybrid.size()!=0);
    pQue->m_vClock = pEvent->pPacket->m_time;
  }
  else // for Hybrid switch
  {
    pQue = SelectQueue(pEvent);
  }
  pQue->m_burstSize += pEvent->pVc->m_burstSize;
  pQue->m_tokenRate += 1.0/pEvent->pVc->m_tokenInterval;
  pQue->m_tick = INT64(1/pQue->m_tokenRate);
  AdjustQueueThreshold();

  pNewVc = m_vcHash.AddVc(pEvent->pVc);
  pQue->m_listPVc.push_back(pNewVc);
  pNewVc->m_pQueHybrid = pQue;
}

// select a HybridQueue for this VC
// different versions of this function should be tested
PQueueHybrid Switch::SelectQueue(PEvent pEvent)
{
  switch(m_method)
  {
  case PRE_SET:   // select a queue with the same id as indicated
    return SelectPreSet(pEvent);
  case RATE_BALANCE:  // select a queue with smallest token rate
    return SelectMinRate(pEvent);
  case BURST_BALANCE:  // select a queue with smallest token rate
    return SelectMinBurst(pEvent);
  case SIMILAR:    // select a queue with similar traffic profile
    return SelectSimilar(pEvent);
  case RANDOM:     // select a random queue
    return SelectRandom(pEvent);
  case ROUND_ROBIN:// select in a sequence of round robin
    return SelectRoundRobin(pEvent);
  case ADAPTIVE:// select in a sequence of round robin
    return SelectAdaptive(pEvent);
  case NEAR_OPTIMAL:    // use optimal algorithm
    return SelectNearOptimal(pEvent);
  case OPTIMAL:    // use optimal algorithm
    return SelectOptimal(pEvent);
  default:
    Error("Error in SelectQueue\n");
  }
  return NULL;
}

PQueueHybrid Switch::SelectMinRate(PEvent pEvent)
{
  VectorPQueueHybridIt it;
  INT64 maxTick;
  PQueueHybrid pQue;

  if (m_vecPQueHybrid.size()<m_maxQueue)
  {
    pQue = new QueueHybrid;
    pQue->m_id = m_vecPQueHybrid.size();
    m_vecPQueHybrid.push_back(pQue);
    assert(m_vecPQueHybrid.size()!=0);
    pQue->m_vClock = pEvent->pPacket->m_time;
  }
  else
  {
    maxTick = 0;
    pQue = NULL;
    for (it=m_vecPQueHybrid.begin(); it != m_vecPQueHybrid.end(); it++)
      if ((*it)->m_tick > maxTick)
      {
        pQue = *it;
        maxTick = (*it)->m_tick;
      }
    if (pQue==NULL)
      pQue = *m_vecPQueHybrid.begin();
  }
  return pQue;
}

PQueueHybrid Switch::SelectMinBurst(PEvent pEvent)
{
  VectorPQueueHybridIt it;
  double minBurst;
  PQueueHybrid pQue;

  if (m_vecPQueHybrid.size()<m_maxQueue)
  {
    pQue = new QueueHybrid;
    pQue->m_id = m_vecPQueHybrid.size();
    m_vecPQueHybrid.push_back(pQue);
    pQue->m_vClock = pEvent->pPacket->m_time;
  }
  else
  {
    minBurst = DBL_MAX;
    pQue = NULL;
    for (it=m_vecPQueHybrid.begin(); it != m_vecPQueHybrid.end(); it++)
      if ((*it)->m_burstSize < minBurst)
      {
        pQue = *it;
        minBurst = pQue->m_burstSize;
      }
    assert(pQue!=NULL);
  }
  return pQue;
}

PQueueHybrid Switch::SelectPreSet(PEvent pEvent)
{
  VectorPQueueHybridIt it;
  PQueueHybrid pQue;

  for (it=m_vecPQueHybrid.begin(); it != m_vecPQueHybrid.end(); it++)
    if ((*it)->m_id == pEvent->pVc->m_qid)
      return *it;
  pQue = new QueueHybrid;
  pQue->m_id = pEvent->pVc->m_qid;
  pQue->m_vClock = pEvent->pPacket->m_time;
  m_vecPQueHybrid.push_back(pQue);
  return pQue;
}

PQueueHybrid Switch::SelectSimilar(PEvent pEvent)
{
  double minDiff;
  PQueueHybrid pQue, pSimilarQue;

  pSimilarQue = SelectSimilarFromRange(pEvent,0,m_vecPQueHybrid.size(),minDiff);
  if (m_vecPQueHybrid.size()<m_maxQueue)
  {
    if (minDiff <= 1e-5)
      return pSimilarQue;
    pQue = new QueueHybrid;
    pQue->m_id = m_vecPQueHybrid.size();
    m_vecPQueHybrid.push_back(pQue);
    pQue->m_vClock = pEvent->pPacket->m_time;
    return pQue;
  }
  else
  {
    assert(pSimilarQue!=NULL);
    return pSimilarQue;
  }
}

PQueueHybrid Switch::SelectRandom(PEvent pEvent)
{
  PQueueHybrid pQue;
  size_t i;
  VectorPQueueHybridIt it;

  if (m_vecPQueHybrid.size()<m_maxQueue)
  {
    pQue = new QueueHybrid;
    pQue->m_id = m_vecPQueHybrid.size();
    m_vecPQueHybrid.push_back(pQue);
    pQue->m_vClock = pEvent->pPacket->m_time;
  }
  else
  {
    i = size_t(double(m_rand.random()) / RAND_M * m_maxQueue);
    for (it=m_vecPQueHybrid.begin(); it != m_vecPQueHybrid.end(); it++)
    {
      pQue = *it;
      if (pQue->m_id == i)
        break;
    }
  }
  return pQue;
}

PQueueHybrid Switch::SelectRoundRobin(PEvent pEvent)
{
  PQueueHybrid pQue;
  VectorPQueueHybridIt it;

  if (m_vecPQueHybrid.size()<m_maxQueue)
  {
    pQue = new QueueHybrid;
    pQue->m_id = m_vecPQueHybrid.size();
    m_vecPQueHybrid.push_back(pQue);
    pQue->m_vClock = pEvent->pPacket->m_time;
  }
  else
  {
    for (it=m_vecPQueHybrid.begin(); it != m_vecPQueHybrid.end(); it++)
    {
      pQue = *it;
      if (pQue->m_id == m_curRRIndex)
        break;
    }
    m_curRRIndex = (m_curRRIndex+1) % m_maxQueue;
  }
  return pQue;
}

PQueueHybrid Switch::SelectAdaptive(PEvent pEvent)
{
  PQueueHybrid pQue, pSimilarQue;
  int bound[4];
  size_t size;
  double minDiff;

  size = m_vecPQueHybrid.size();
  bound[0] = 0;
  bound[1] = round(size*1.0/3);
  bound[2] = round(size*2.0/3);
  bound[3] = size;
  pSimilarQue = SelectSimilarFromRange(pEvent,bound[pEvent->pVc->m_type],
                                    bound[pEvent->pVc->m_type+1],minDiff);

  if (size<m_maxQueue)
  {
    if (minDiff <= 1e-5)
      return pSimilarQue;
    pQue = new QueueHybrid;
    pQue->m_id = size;
    m_vecPQueHybrid.push_back(pQue);
    pQue->m_vClock = pEvent->pPacket->m_time;
    return pQue;
  }
  else
  {
    verify(pSimilarQue!=NULL);
    return pSimilarQue;
  }
}

PQueueHybrid Switch::SelectOptimal(PEvent pEvent)
{
  size_t i, size, iteration, MaxIteration;
  double benifit, prebenifit, diff, MinDiff;
  PQueueHybrid pQue;
  VectorQueue vecQue;
  Queue que, *pque;
  Flow flow, *pflow;
  ListPVCIt pVcIt;

  if (m_vecPQueHybrid.size()<m_maxQueue)
  {
    pQue = new QueueHybrid;
    pQue->m_id = m_vecPQueHybrid.size();
    m_vecPQueHybrid.push_back(pQue);
    pQue->m_vClock = pEvent->pPacket->m_time;
    return pQue;
  }
  else
  {
    MaxIteration = 50;
    MinDiff = 1e-8;
    prebenifit = 0;
    size = m_vecPQueHybrid.size();
    for (i=0; i<size; i++)
    {
      vecQue.push_back(que);
      pque = &(vecQue.back());
      pque->m_burst = m_vecPQueHybrid[i]->m_burstSize;
      pque->m_rate = m_vecPQueHybrid[i]->m_tokenRate;
      for (pVcIt = m_vecPQueHybrid[i]->m_listPVc.begin();
           pVcIt != m_vecPQueHybrid[i]->m_listPVc.end();
           pVcIt++)
      {
        pque->m_listFlow.push_back(flow);
        pflow = &(pque->m_listFlow.back());
        pflow->m_id = (*pVcIt)->m_id;
        pflow->m_burst = (*pVcIt)->m_burstSize;
        pflow->m_rate = 1.0/(*pVcIt)->m_tokenInterval;
      }
    }
    for (iteration=0; iteration<MaxIteration; iteration++)
    {
      // for all VC pVc
        // for all Queue get max benifit
      benifit = 0;
      diff = benifit-prebenifit;
      if (diff < MinDiff)
        break;
    } //for (iteration=0; iteration<MaxIteration; iteration++)

    return NULL;
  }
}

PQueueHybrid Switch::SelectSimilarFromRange(PEvent pEvent, 
                                            int begin, int end, double &minDiff)
{
  double diff;
  VectorPQueueHybridIt it;
  PQueueHybrid pQue, pMinQue;
  double burstSize;
  double tokenRate, avgBurstSize, avgTokenRate;

  burstSize = pEvent->pVc->m_burstSize;
  tokenRate = 1.0/pEvent->pVc->m_tokenInterval;
  minDiff = DBL_MAX;
  pMinQue = NULL;
  for (it=m_vecPQueHybrid.begin(); it != m_vecPQueHybrid.end(); it++)
  {
    pQue = *it;
    if ((pQue->m_id<begin)||(pQue->m_id>=end))
      continue;
    if (pQue->m_listPVc.size()==0)
    {
      avgBurstSize = 0.0;
      avgTokenRate = 0.0;
      diff = 2;
    }
    else
    {
      avgBurstSize = double(pQue->m_burstSize)/pQue->m_listPVc.size();
      avgTokenRate =  pQue->m_tokenRate/pQue->m_listPVc.size();
      diff = fabs((double(burstSize) - avgBurstSize)/avgBurstSize)
             + fabs((tokenRate - avgTokenRate)/avgTokenRate);
    }
    if (diff < minDiff)
    {
      minDiff=diff;
      pMinQue = pQue;
    }
  }
  return pMinQue;
}

PQueueHybrid Switch::SelectNearOptimal(PEvent pEvent)
{
  size_t i, size;
  int queue;
  double maxbenifit, benifit;
  PQueueHybrid pQue;

  if (m_vecPQueHybrid.size()<m_maxQueue)
  {
    pQue = new QueueHybrid;
    pQue->m_id = m_vecPQueHybrid.size();
    m_vecPQueHybrid.push_back(pQue);
    pQue->m_vClock = pEvent->pPacket->m_time;
    return pQue;
  }
  else
  {
    size = m_vecPQueHybrid.size();
    queue = -1;
    maxbenifit = 0.0;
    for (i=0; i<size; i++)
    {
      benifit = GetOptBenifit(i, pEvent->pVc->m_burstSize, 1.0/pEvent->pVc->m_tokenInterval);
      if (benifit > maxbenifit)
      {
        maxbenifit = benifit;
        queue = i;
      }
    }
    if (queue==-1)
      queue=0;
    return m_vecPQueHybrid[queue];
  }
}

double Switch::GetOptBenifit(size_t k, double burst, double rate)
{
  size_t i, j, size;
  double cost, d, b1, r1, b2, r2;

  size = m_vecPQueHybrid.size();
  cost = 0.0;
  for (i=0; i<size; i++)
  {
    b1 = m_vecPQueHybrid[i]->m_burstSize;
    r1 = m_vecPQueHybrid[i]->m_tokenRate;
    if (i==k)
    {
      b1 += burst;
      r1 += rate;
    }
    for (j=i+1; j<size; j++)
    {
      b2 = m_vecPQueHybrid[j]->m_burstSize;
      r2 = m_vecPQueHybrid[j]->m_tokenRate;
      if (j==k)
      {
        b2 += burst;
        r2 += rate;
      }
      d = sqrt(b1*r2)-sqrt(b2*r1);
      cost += d*d;
    }
  }
  return cost;
}

// if q1 is later than q2, return TRUE;
BOOL Switch::LaterVirtualClock(PQueueHybrid &q1, PQueueHybrid &q2)
{
  if (q1->m_quePPacket.empty())
  {
    if (q2->m_quePPacket.empty())
    {
      if (q1->m_id > q2->m_id)
        return TRUE;
      else
        return FALSE;
    }
    else
      return TRUE;
  }
  else // q1 is not empty
  {
    if (q2->m_quePPacket.empty())
      return FALSE;
    else
    {
      if (q1->m_vClock > q2->m_vClock)
        return TRUE;
      else if (q1->m_vClock < q2->m_vClock)
        return FALSE;
      else // ==
      {
        if (q1->m_id > q2->m_id)
          return TRUE;
        else
          return FALSE;
      }
    }
  }
}

// reallocate the buffer size of each queue
void Switch::AdjustQueueThreshold()
{
  VectorPQueueHybridIt ppQueIt;
  PQueueHybrid pQue;
  VectorDouble vecThres;
  double thres, totalThres;
  DWORD threshold, total;
  DWORD i, size;

  VectorDouble vecAlpha;
  double rate, totalRate, alpha, totalAlpha;

  size = m_vecPQueHybrid.size();
  if ((m_switchType == Hybrid)||(m_switchType == WFQ))
  {
    totalRate = 0.0;
    totalAlpha = 0.0;
    for (ppQueIt=m_vecPQueHybrid.begin();
         ppQueIt!=m_vecPQueHybrid.end();
         ppQueIt++)
    {
      pQue = *ppQueIt;
      alpha = sqrt(pQue->m_burstSize * pQue->m_tokenRate);
      vecAlpha.push_back(alpha);
      totalAlpha += alpha;
      totalRate += pQue->m_tokenRate;
    }
    totalThres = 0.0;
    for (i = 0; i < size; i++)
    {
      pQue = m_vecPQueHybrid[i];
      assert(1.0/m_outputInterval>=totalRate);
      if (totalAlpha<1e-11)
      {
          pQue->m_tick = UINT_MAX;
          thres = 0.0;
      }
      else
      {
        rate = pQue->m_tokenRate + 
               double(vecAlpha[i])/totalAlpha*(1.0/m_outputInterval - totalRate);
        assert(rate>=0);
        if (rate < 1e-11)
        {
          pQue->m_tick = UINT_MAX;
          thres = 0.0;
        }
        else
        {
          pQue->m_tick = 1/rate;
          thres = rate * pQue->m_burstSize / ( rate - pQue->m_tokenRate);
          assert(thres>=0);
        }
      }
      vecThres.push_back(thres);
      totalThres += thres;
    }

    total = 0;
    for (i = 0; i < size; i++)
    {
      pQue = m_vecPQueHybrid[i];
      threshold = DWORD(vecThres[i]/totalThres*m_bufSize);
      m_vecPQueHybrid[i]->m_threshold = threshold;
      totalThres += threshold;
      total += threshold;
      /*
      if (threshold < m_vecPQueHybrid[i]->m_quePPacket.size())  // we need drop packets
      {
        dropNum = m_vecPQueHybrid[i]->m_quePPacket.size() - threshold;
        for (k=0; k<dropNum; k++)
        {
          pVc = OutputPacket(m_vecPQueHybrid[i]);
          if (pVc!=NULL)
          {
            pVc->m_allLoss++;
            m_vecPQueHybrid[i]->m_allLoss++;
            m_allLoss++;
            if (m_curTime > m_warmUp)
            {
              pVc->m_loss++;
              m_vecPQueHybrid[i]->m_loss++;
              m_loss++;
            }
          }
          else if (m_vecPQueHybrid[i]->m_listPVc.size()==0) // erase this queue
          {
            assert(0);
            Error("Internal error in AdjustQueueThreshold!\n");
          } 
        }
      }
      */
    }
    for (i=0; total<m_bufSize;i = (i+1)%size)
    {
      if (m_vecPQueHybrid[i]->m_listPVc.size()!=0)
      {
        total++;
        m_vecPQueHybrid[i]->m_threshold++;
      }
    }
  }
  else
  {
    totalThres = 0.0;
    for (ppQueIt=m_vecPQueHybrid.begin();
         ppQueIt!=m_vecPQueHybrid.end();
         ppQueIt++)
    {
      thres = (*ppQueIt)->m_burstSize + 
              (*ppQueIt)->m_tokenRate * m_bufSize * m_outputInterval;
      vecThres.push_back(thres);
      totalThres += thres;
    }
    total = 0;
    for (i = 0; i < size; i++)
    {
      threshold = DWORD(double(vecThres[i])/totalThres*m_bufSize);
      m_vecPQueHybrid[i]->m_threshold = threshold;
      total += threshold;
      /*
      if (threshold < m_vecPQueHybrid[i]->m_quePPacket.size())  // we need drop packets
      {
        dropNum = m_vecPQueHybrid[i]->m_quePPacket.size() - threshold;
        for (k=0; k<dropNum; k++)
        {
          pVc = OutputPacket(m_vecPQueHybrid[i]);
          if (pVc!=NULL)
          {
            pVc->m_allLoss++;
            m_vecPQueHybrid[i]->m_allLoss++;
            m_allLoss++;
            if (m_curTime > m_warmUp)
            {
              pVc->m_loss++;
              m_vecPQueHybrid[i]->m_loss++;
              m_loss++;
            }
          }
          else if (m_vecPQueHybrid[i]->m_listPVc.size()==0) // erase this queue
          {
            assert(0);
            Error("Internal error in AdjustQueueThreshold!\n");
          } 
        }
      }
      */
    } // for
    for (i=0; total<m_bufSize;i = (i+1)%size)
    {
      if (m_vecPQueHybrid[i]->m_listPVc.size()!=0)
      {
        total++;
        m_vecPQueHybrid[i]->m_threshold++;
      }
    }
  }
}

// update all virtual time to current time
void Switch::CheckVClock(INT64 curTime)
{
  if ((m_switchType != FIFO_BUF) && (m_switchType != FIFO))
  {
    if (curTime - m_checkTime > m_interval)
    {
      VectorPQueueHybridIt it;

      m_checkTime = curTime;
      for (it=m_vecPQueHybrid.begin(); it != m_vecPQueHybrid.end(); it++)
        if ((*it)->m_vClock<curTime)
          (*it)->m_vClock = curTime;
      make_heap(m_vecPQueHybrid.begin(), m_vecPQueHybrid.end(), LaterVirtualClock);
    }
  }
}

void Switch::Check()
{
  PQueueHybrid pQue;
  VectorPQueueHybridIt ppQueIt;
  DWORD loss, outNum, total, totalInBuf;
  size_t vcnum;
  double rate;

  loss=outNum=total=totalInBuf=0;
  rate = 0.0;
  vcnum=0;
  for (ppQueIt=m_vecPQueHybrid.begin(); 
       ppQueIt != m_vecPQueHybrid.end(); 
       ppQueIt++)
  {
    pQue = *ppQueIt;
    pQue->Check(vcnum);
    rate += pQue->m_tokenRate;
    loss += pQue->m_allLoss;
    outNum += pQue->m_allOutNum;
    total += pQue->m_allTotal;
    totalInBuf += pQue->m_quePPacket.size();
    if ((m_switchType==FIFO)||(m_switchType==FIFO_BUF))
    {
      //assert(pQue->m_quePPacket.size()<=pQue->m_threshold);
    }
  }
  //verify(vcnum<=30);
  //verify(rate <= 0.00000866);
  verify(rate < 1.0/m_outputInterval);
  verify(total == m_allTotal);
  verify(total == loss + outNum + totalInBuf);
  verify(totalInBuf <= m_bufSize);
}

void Switch::PrintThreshold()
{
  PQueueHybrid pQue;
  VectorPQueueHybridIt ppQueIt;
  QueuePPacketIt pQuePPacketIt;
  int i;

  cout << m_name << " Id\tThreshold\tTick\n";
  for (i=0; i<m_vecPQueHybrid.size(); i++)
  {
    for (ppQueIt=m_vecPQueHybrid.begin(); 
         ppQueIt != m_vecPQueHybrid.end(); 
         ppQueIt++)
    {
      pQue = *ppQueIt;
      if (pQue->m_id!=i)
        continue;
      cout << pQue->m_id << '\t' << pQue->m_threshold 
           << '\t' << pQue->m_tick << endl;
    }
  }
  TRACE("\n");
}

void Switch::TraceAll()
{
  return;
  PQueueHybrid pQue;
  VectorPQueueHybridIt ppQueIt;
  QueuePPacketIt pQuePPacketIt;
  PPacket pPacket;
  int i;

  TRACE("%s %I64d id threshold[length] tick:[all packets]\n",m_name, m_curTime);
  for (i=0; i<m_vecPQueHybrid.size(); i++)
  {
    for (ppQueIt=m_vecPQueHybrid.begin(); 
         ppQueIt != m_vecPQueHybrid.end(); 
         ppQueIt++)
    {
      pQue = *ppQueIt;
      if (pQue->m_id!=i)
        continue;
      TRACE("id:%d %d[%d] %d:", pQue->m_id, pQue->m_threshold,
            pQue->m_quePPacket.size(), pQue->m_tick);
      for (pQuePPacketIt=pQue->m_quePPacket.begin();
           pQuePPacketIt!=pQue->m_quePPacket.end();pQuePPacketIt++)
      {
        pPacket = *pQuePPacketIt;
        TRACE("(%d,%d)",pPacket->m_vcId,pPacket->m_seq);
      }
      TRACE("\n");
    }
  }
  TRACE("\n");
}

void Switch::DebugPrint()
{
  VectorPQueueHybridIt itQue;
  PQueueHybrid pQue;
  ListPVCIt itPVc;
  PVC pVc;
  char s[100];
  double peak, avg;
  
  cout << m_name << endl;
  //TRACE("%I64d\n",m_curTime);
  for (itQue=m_vecPQueHybrid.begin(); itQue != m_vecPQueHybrid.end(); itQue++)
  {
    pQue = *itQue;
    cout << pQue->m_id << ":";
    for (itPVc=pQue->m_listPVc.begin(); itPVc!= pQue->m_listPVc.end(); itPVc++)
    {
      pVc = *itPVc;
      peak = double(TimeScale)*8.0*CELL_SIZE / pVc->m_peakInterval / 1e6;
      avg = pVc->m_avgRate / (1e6 / 8.0 / double(CELL_SIZE) / double(TimeScale));
      sprintf(s, " (%3.1lf,%3.1lf)", peak, avg);
      cout << s;
    }
    //sprintf(s, " Loss: %8.5lf%%\n",pQue->m_allLoss*1.0/pQue->m_allTotal*100);
    //cout << " Thres:" << pQue->m_threshold << " Tick:" << pQue->m_tick
    //     << " Len:" << pQue->m_quePPacket.size() 
    //     << s;
    cout << endl;
    sprintf(s, "%d: Loss: %8.5lf%%\n",pQue->m_id,
                pQue->m_allLoss*1.0/pQue->m_allTotal*100);
    //TRACE(s);
  }
  cout << endl << flush;
}

ostream& operator<<( ostream& os, QueueHybrid &q )
{
  return os;
}

ostream& operator<<( ostream& os, Switch &s )
{
  char str[127];
  double throughput;
  double throughputs[3], loss[3], total[3], maxloss[3];
  double lossratio;
  PQueueHybrid pQue;
  ListPVCIt pVcIt;
  PVC pVc;
  VectorPQueueHybridIt ppQueIt;
  int type;
  int i;
  double excess[2];

  if (s.m_curTime>s.m_warmUp)
  {
    for (i=0; i<3; i++)
    {
      loss[i] = 0.0;
      total[i] = 0.0;
      maxloss[i]=0.0;
    }
    for (ppQueIt=s.m_vecPQueHybrid.begin(); 
         ppQueIt != s.m_vecPQueHybrid.end(); 
         ppQueIt++)
    {
      pQue = *ppQueIt;
      os << *pQue;

      for (pVcIt = pQue->m_listPVc.begin();
           pVcIt != pQue->m_listPVc.end(); 
           pVcIt++)
      {
        pVc = *pVcIt;
        type = pVc->m_type;
        loss[type] += pVc->m_loss;
        total[type] += pVc->m_total - pVc->m_num;
        lossratio = double(pVc->m_loss) / (pVc->m_total - pVc->m_num);
        if (lossratio>maxloss[type])
          maxloss[type] = lossratio;
        if (pVc->m_id==0)
          excess[0] = double(pVc->m_total - pVc->m_num - pVc->m_loss) * double(CELL_SIZE) / 
                      (s.m_curTime-s.m_warmUp) * 8e3;
        else if (pVc->m_id==3)
          excess[1] = double(pVc->m_total - pVc->m_num - pVc->m_loss) * double(CELL_SIZE) / 
                      (s.m_curTime-s.m_warmUp) * 8e3;
      }
      for (pVcIt = pQue->m_listPVcPast.begin();
           pVcIt != pQue->m_listPVcPast.end(); 
           pVcIt++)
      {
        pVc = *pVcIt;
        type = pVc->m_type;
        loss[type] += pVc->m_loss;
        total[type] += pVc->m_total - pVc->m_num;
      }
    }

    throughput = double(s.m_outNum) * CELL_SIZE / (s.m_curTime-s.m_warmUp) * 8e3;
    for (i=0; i<3; i++)
      throughputs[i] = (total[i]-loss[i]) * double(CELL_SIZE) / (s.m_curTime-s.m_warmUp) * 8e3;

    //Name        Buffer  Headrm Throu AggrThrou ConfThrou ConfLoss  ModThrou  ModLoss
    //            MBytes  MBytes Mb/s  Mb/s       Mb/s               Mb/s
    sprintf(str, "%-12s %5.3lf %5.3lf %6.3lf %6.3lf %6.3lf %6.3lf%% %6.3lf %6.3lf%% %lf %6.3lf %6.3lf\n", 
            s.m_name, s.m_bufSize*CELL_SIZE/1e6, s.m_totalHeadroom*CELL_SIZE/1e6,
            throughput, throughputs[2], throughputs[0], 
            total[0]==0?0:double(loss[0])*100.0/total[0],
            throughputs[1], total[1]==0?0:double(loss[1])*100.0/total[1], s.GetOptBenifit(), 
            excess[0], excess[1]); // throughput of flow 6 and 8
  }
  else
  {
    //Name        Buffer  Headroom  Throu AggrThrou ConfThrou ConfLoss
    //            MBytes  MBytes    Mb/s    Mb/s       Mb/s             
    sprintf(str, "%-13s %7.4lf %7.4lf warming up\n", 
            s.m_name, s.m_bufSize*CELL_SIZE/1e6, s.m_totalHeadroom*CELL_SIZE/1e6);
  }

  os << str;

  return os;
}
