// VC.cpp: implementation of the VC class.
//
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "Event.h"
#include "VC.h"

int VC::m_seq = 0;         // sequence number of packets

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

VC::VC(int id, int qid,
       double beginTimeSec, double lengthSec, // seconds
       double peakRate, double avgRate, double tokenRate, // Mbits/s
       double tokenBucket, // KBytes
       int type,
       double onTime, // seconds
       int seedOn, int seedOff)
{
  m_id = id;
  m_qid = qid;

  m_allOutNum = 0;
  m_outNum = 0;
  m_num = 0;     // current number of packets in the output queue
  m_loss = 0;    // number of packets that are lost
  m_allLoss = 0;
  m_pQueHybrid = NULL; // the pointer to the HybridQueue which contains this vc
  m_flag = VC_NORMAL;    // indicate whether the vc stops
  m_total = 0;   // total packets of this vc until now
  m_allTotal = 0;   // total packets of this vc until now

  m_beginTime = beginTimeSec * TimeScale;    // begin time of this vc in ns
  m_endTime = m_beginTime + lengthSec*TimeScale;      // end time of this vc in ns

  m_peakInterval = DWORD(TimeScale*8.0*CELL_SIZE / peakRate / 1e6);
  m_tokenBucket = DWORD(tokenBucket * 1e3 / CELL_SIZE);
  m_tokenInterval = DWORD(TimeScale*8.0*CELL_SIZE / tokenRate / 1e6); // token rate
  m_avgRate = double(avgRate)*1e6 / 8.0 / CELL_SIZE / TimeScale;
  m_burstSize = m_tokenBucket; // * peakRate / (peakRate - tokenRate);
  m_type = type;
  switch (m_type)
  {
  case 0: // modulated conform
    m_conform = TRUE;
    break;
  case 1: // non-regulate
    m_conform = FALSE;
    break;
  case 2: // aggressive non-regulate
    m_conform = FALSE;
    break;
  default:
    Error("Wrong type!");
  }

  if (m_conform)
    m_avgOn = m_burstSize*m_peakInterval; // test a smaller avgOn
  else
    m_avgOn = INT64(onTime*TimeScale);
  m_avgOff = INT64(m_avgOn / (m_avgRate * m_peakInterval) - m_avgOn);
  m_randOn.srandom(seedOn);
  m_randOff.srandom(seedOff);

  m_onEndTime = m_beginTime; // begin from an OFF state
  m_onBeginTime = m_beginTime+OffPeriod();
  m_on = FALSE;
  //m_onEndTime = m_beginTime+OnPeriod(); // begin from an ON state
  //m_onBeginTime = m_beginTime;
  //m_on = TRUE;
  m_prevCell = m_beginTime;     // arrival time of the current cell
  m_prevToken = m_beginTime;
  m_prevTime = m_beginTime;
  m_curBucket = m_tokenBucket;
  m_prevEvt = EVT_NOTHING;
  m_waiting = 0;
}

VC::~VC()
{

}

PEvent VC::GetNextEvent()
{
  PEvent pevt;

  assert(m_prevEvt!=EVT_VC_STOP);

  pevt = new Event;
  if (m_prevEvt==EVT_NOTHING) // this is the first event
  {
    pevt->what = EVT_VC_START;
    pevt->pPacket = new Packet(m_seq++, m_id, m_beginTime);
    pevt->pVc = new VC(*this);
    m_prevEvt = pevt->what;

    m_prevTime = pevt->pPacket->m_time;
    return pevt;
  } 
  else if (m_prevCell+m_peakInterval >= m_endTime) // this vc should stop
  {
    pevt->what = EVT_VC_STOP;
    pevt->pPacket = new Packet(m_seq++, m_id, m_prevTime);
    m_prevEvt = pevt->what;

    assert(m_prevTime <= pevt->pPacket->m_time);
    m_prevTime = pevt->pPacket->m_time;
    return pevt;
  }
  else
  {
    INT64 numToken;
    if (m_conform) // use leaky bucket (infinite cell buffer and token pool)
    {
      while (TRUE)
      {
        if (m_on) // on period
        {
          m_prevCell += m_peakInterval; // a new cell may arrive at this time
          if (m_prevCell > m_onEndTime) // if it is out of the ON period
          {
            m_onBeginTime = m_onEndTime + OffPeriod();
            m_on = FALSE;
            continue;
          }

          numToken = (m_prevCell - m_prevToken) / m_tokenInterval;
          assert(numToken>=0);
          if (numToken>0) // add tokens to the token pool
          {
            m_curBucket += DWORD(numToken);
            if (m_curBucket>m_tokenBucket)
              m_curBucket = m_tokenBucket;
            m_prevToken += numToken * m_tokenInterval;
          }
          if (m_curBucket>0) // there is token avaiable
          {                   // return this cell
            if (m_prevCell >= m_endTime)
            {
              pevt->what = EVT_VC_STOP;
            }
            else
              pevt->what = EVT_PACKET;
            m_curBucket--;
            pevt->pPacket = new Packet(m_seq++, m_id, m_prevCell);
            m_prevEvt = pevt->what;
            assert(m_prevTime <= pevt->pPacket->m_time);
            m_prevTime = pevt->pPacket->m_time;
            return pevt;
          }
          else
            m_waiting++;
        }
        else // off period
        {
          m_prevToken += m_tokenInterval; // a new cell may arrive at this time
          if (m_prevToken > m_onBeginTime) // if it is out of the OFF period
          {
            m_prevToken -= m_tokenInterval;
            m_onEndTime = m_onBeginTime + OnPeriod();
            m_prevCell = m_onBeginTime - m_peakInterval;
            m_on = TRUE;
            continue;
          }
          if (m_waiting > 0)
          {
            m_waiting--;
            pevt->what = EVT_PACKET;
            pevt->pPacket = new Packet(m_seq++, m_id, m_prevToken);
            m_prevEvt = pevt->what;
            assert(m_prevTime <= pevt->pPacket->m_time);
            m_prevTime = pevt->pPacket->m_time;
            return pevt;
          }
          else
          {
            m_curBucket++;
            if (m_curBucket>m_tokenBucket)
              m_curBucket = m_tokenBucket;
          }
        }
      }
      /* this is the buffered leaky bucket
      int endTime;
      INT64 nextOnTime;

      while (TRUE)  // while no cell arrives, advance time until get one cell
      {
        m_prevToken += m_tokenInterval;     // generate a token
        endTime = m_prevToken<m_onEndTime ? m_prevToken : m_onEndTime;
        // cells generated before the token
        numCells = (endTime - m_prevCell)/m_peakInterval;
        assert(numCells != 0);
        //if (numCells == 0)  // if no cells come in
        //  continue;         // continue advance time

        m_prevCell += numCells * m_peakInterval;
        if (m_prevCell > m_onEndTime)
        {
          nextOnTime = m_onEndTime + OffPeriod();
          m_prevCell = nextOnTime - m_peakInterval; // So next cell will start on nextOnTime
          m_onEndTime = nextOnTime + OnPeriod();
          m_prevToken += (nextOnTime - m_prevToken)/m_tokenInterval*m_tokenInterval;
        }

        // put the cells into the token bucket
        m_curBucket -= numCells;  
        if (m_curBucket < 0)
          m_curBucket = 0;

        if (m_curBucket < m_tokenBucket) // there is cells in the bucket
        {
          m_curBucket++;
          pevt->what = EVT_PACKET;
          pevt->pPacket = new Packet(m_seq++, m_id, m_prevToken);
          m_prevEvt = pevt->what;
          return pevt;
        }
      }
      */
    }
    else
    { // do not use leaky bucket
      m_prevCell += m_peakInterval;
      if (m_prevCell > m_onEndTime)
      {
        m_prevCell = m_onEndTime + OffPeriod();
        m_onEndTime = m_prevCell + OnPeriod();
      }
      if (m_prevCell >= m_endTime)
      {
        pevt->what = EVT_VC_STOP;
        m_prevCell = m_endTime;
      }
      else
        pevt->what = EVT_PACKET;
      pevt->pPacket = new Packet(m_seq++, m_id, m_prevCell);
      m_prevEvt = pevt->what;
      assert(m_prevTime <= pevt->pPacket->m_time);
      m_prevTime = pevt->pPacket->m_time;
      return pevt;
    }
  }
  assert(FALSE);
  return NULL;
}

INT64 VC::NextPossionEvent(RandGen &randGen, DWORD avgInterval)
{
  double rand;
  double r;

  rand = double(randGen.random()) / RAND_M;
  r = -log(rand)*avgInterval;
  assert(r>=0);
  return INT64(r);
}

INT64 VC::OnPeriod()
{ 
  if (m_conform)
    return m_avgOn;
  else
    return NextPossionEvent(m_randOn,DWORD(m_avgOn)); 
}

INT64 VC::OffPeriod()
{ 
  //if (m_conform)
  //  return m_avgOff;
  //else
    return NextPossionEvent(m_randOff,DWORD(m_avgOff)); 
}

void VC::SetSeed(int seedOn, int seedOff)
{
  int i;
  m_randOn.srandom(seedOn);
  m_randOff.srandom(seedOff);
  for (i=0; i<100; i++)
  {
    m_randOn.random();
    m_randOff.random();
  }
}
